/// <reference types="google.maps" />
import { Autocomplete, Circle, Cluster, DrawingManager, HeatmapLayer, InfoWindow, KmlLayer, MapLayer, Marker, Polygon, Polyline, Rectangle, StreetViewPanorama } from '../components';
import type { IGmapVueElementOptions, ILoadPluginOptions } from '../interfaces';
import type { Emitter, EventType } from 'mitt';
import type { ComponentOptions, Ref } from 'vue';
export type GoogleMapsAPIInitializerFn = (options: ILoadPluginOptions, loadCn?: boolean) => void;
export type GmvComponents = {
    Autocomplete: typeof Autocomplete;
    Circle: typeof Circle;
    Cluster: typeof Cluster;
    DrawingManager: typeof DrawingManager;
    HeatmapLayer: typeof HeatmapLayer;
    InfoWindow: typeof InfoWindow;
    KmlLayer: typeof KmlLayer;
    MapLayer: typeof MapLayer;
    Marker: typeof Marker;
    Polyline: typeof Polyline;
    Polygon: typeof Polygon;
    Rectangle: typeof Rectangle;
    StreetViewPanorama: typeof StreetViewPanorama;
};
export type GmvSharedComposables = {
    useMapPromise: () => Promise<google.maps.Map | undefined>;
    useResizeBus: () => {
        currentResizeBus: Ref<Emitter<Record<EventType, unknown>> | undefined>;
        _resizeCallback: () => void;
        _delayedResizeCallback: () => Promise<void>;
    };
    useGoogleMapsApiPromiseLazy: () => Promise<any>;
    useStreetViewPanoramaPromise: () => Promise<google.maps.StreetViewPanorama | undefined>;
};
export type GmvUtilities = {
    googleMapsApiInitializer: GoogleMapsAPIInitializerFn;
    pluginComponentBuilder: (providedOptions: IGmapVueElementOptions) => ComponentOptions;
    getGoogleMapsAPI: () => false | typeof google;
};
